from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import numpy
from PyMca5.PyMcaMath.linalg import lstsq


def fit_profiles(
    base_profiles: numpy.ndarray,
    profiles: numpy.ndarray,
    positive_constraint: Sequence[int] = (),
    weight: bool = True,
    profiles_stdev: Optional[numpy.ndarray] = None,
) -> Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]:
    """Solve linear system of equations `A.p = b` for several profiles.
    Some parameters might be have `>=0` constraints. This implementation
    use the `lstsq` solver from PyMca.

    :param base_profiles: `A` with shape ``(num_channels, num_parameters)``
    :param profiles: `b` with shape ``(num_channels, num_profiles)``
    :param positive_constraint: indices of parameters that should be constraint to `>=0`.
    :param weight: use weights (all ones otherwise).
    :param profiles_stdev: shape ``(num_channels,)`` or ``(num_channels, num_profiles)``
    :returns: parameters `p` with shape ``(num_parameters, num_profiles)``,
              standard deviations on `p` with shape ``(num_parameters, num_profiles)``,
              number of free parameters with shape ``(num_profiles,)``
    """
    num_channels, num_profiles = profiles.shape
    if not weight:
        profiles_stdev = None

    if profiles_stdev is not None:
        if profiles_stdev.ndim == 1:
            profiles_stdev = profiles_stdev[:, numpy.newaxis]
        if profiles_stdev.shape not in (
            (num_channels, num_profiles),
            (num_channels, 1),
        ):
            raise ValueError("Profile and profile STDEV shape is not compatible")

    if len(base_profiles) != num_channels:
        raise ValueError(
            "Profiles and base-profiles must have the same number of observations"
        )

    # SVD or matrix inversion
    matrix_inversion = weight and profiles_stdev is None
    svd = not matrix_inversion

    result = lstsq(
        base_profiles,
        profiles,
        digested_output=True,
        weight=weight,
        sigma_b=profiles_stdev,
        svd=svd,
    )

    parameters = result["parameters"]
    uncertainties = result["uncertainties"]
    num_parameters, _ = parameters.shape  # shape (num_parameters, num_profiles)
    num_free_parameters = numpy.array([num_parameters] * num_profiles)
    if not positive_constraint:
        return parameters, uncertainties, num_free_parameters

    nrefit_max = 2 * len(positive_constraint) + 1
    for refit_count in range(1, nrefit_max + 2):
        # Detect negative values for parameters with a positive constraint
        refit_parameters_indices, profiles_to_fit, negative_list = _refit_info(
            parameters, positive_constraint
        )

        # Re-fit profiles without the parameters that are negative so the parameters
        # that are negative are effectively forced to 0.
        num_profiles_refit = profiles_to_fit.sum() if profiles_to_fit is not None else 0
        num_profiles_min = int(num_profiles / 400)
        do_refit = (
            num_profiles_refit  # no negative parameters
            and num_profiles_refit >= num_profiles_min  # not with refitting few spectra
            and refit_count <= nrefit_max  # probably infinite loop
        )
        if do_refit:
            result = lstsq(
                base_profiles[:, refit_parameters_indices],
                profiles[:, profiles_to_fit],
                digested_output=True,
                weight=weight,
                sigma_b=profiles_stdev,
                svd=svd,
            )

            # Replace re-fitted parameters
            for fit_param_idx, parametersi, uncertaintiesi in zip(
                refit_parameters_indices, result["parameters"], result["uncertainties"]
            ):
                parameters[fit_param_idx, profiles_to_fit] = parametersi
                uncertainties[fit_param_idx, profiles_to_fit] = uncertaintiesi

        # Fix negative parameters to zero
        if negative_list:
            num_free_parameters[:] = num_parameters
        for negative_param_idx, profile_negative_mask in negative_list:
            parameters[negative_param_idx, profile_negative_mask] = 0
            uncertainties[negative_param_idx, profile_negative_mask] = 0
            num_free_parameters[profile_negative_mask] -= 1

        if not do_refit:
            break

    return parameters, uncertainties, num_free_parameters


def _refit_info(
    parameters: numpy.ndarray, positive_constraint: Sequence[int]
) -> Tuple[List[int], Optional[numpy.ndarray], List[Tuple[int, numpy.ndarray]]]:
    """
    :param parameters: shape ``(num_parameters, num_profiles)``
    :param positive_constraint: indices of parameters that should be constraint to `>=0`.
    :returns: indices of free parameters, profile indices to be re-fitted, negative profiles (mask) per parameter
    """
    negative_list: List[Tuple[int, numpy.ndarray]] = []
    # parameter index, profiles (mask) where the parameter is negative

    negative_mask = parameters[positive_constraint] < 0
    for negative_param_idx, profile_negative_mask in zip(
        positive_constraint, negative_mask
    ):
        if profile_negative_mask.any():
            negative_list.append((negative_param_idx, profile_negative_mask))

    if not negative_list:
        return list(positive_constraint), None, negative_list

    negative_list.sort(reverse=True)

    fixed_at_zero_indices = [negative_list[0][0]]
    _, profiles_to_fit = negative_list[0]

    for negative_param_idx, profile_negative_mask in negative_list[1:]:
        combined_mask = profiles_to_fit & profile_negative_mask
        if combined_mask.any():
            fixed_at_zero_indices.append(negative_param_idx)
            profiles_to_fit = combined_mask

    refit_parameters_indices = [
        i for i in range(len(parameters)) if i not in fixed_at_zero_indices
    ]

    return refit_parameters_indices, profiles_to_fit, negative_list



import math
from typing import Tuple


def round_to_significant(value: float, tolerance: float = 0.01) -> float:
    """
    Rounds a number to the least amount of significant decimals without
    changing the value by more than the given tolerance (default 1%).
    """
    if value == 0 or not math.isfinite(value):
        return value
    max_change = tolerance * abs(value)
    decimal_places = -int(math.floor(math.log10(max_change)))
    return round(value, decimal_places)


def round_range(start: float, stop: float, step: float) -> Tuple[float, float, int]:
    """
    Adjust the data range so that the number of bins is an integral number.

    :params start: minimum value
    :params stop: maximum value
    :param step: bin width
    :returns: minimum value, maximum value, number of bins
    """
    if step == 0 or not math.isfinite(start) or not math.isfinite(stop):
        return start, stop, 0
    data_range = stop - start
    nbins = int(round(abs(data_range / step)))
    rounded_range = nbins * step
    adjust = (rounded_range - data_range) / 2
    return start - adjust, stop + adjust, nbins



from typing import Optional

import numpy


class AbstractDataHandler:
    def add_points(self, value: numpy.ndarray) -> None:
        raise NotImplementedError


class AbstractOutputHandler:
    def __enter__(self) -> "AbstractOutputHandler":
        raise NotImplementedError

    def __exit__(self, *args) -> None:
        raise NotImplementedError

    def create_group(self, name: str, data: dict) -> None:
        raise NotImplementedError

    def create_nxdata_handler(
        self,
        group: str,
        name: str,
        npoints: int,
        attrs: Optional[dict] = None,
    ) -> AbstractDataHandler:
        raise NotImplementedError



from contextlib import contextmanager
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy
from PyMca5.PyMcaPhysics.xrf.ClassMcaTheory import McaTheory

from ...xrffit.pymca_config import PyMcaXrfConfiguration
from ...xrffit.pymca_config import pymca_configdict_from_model
from ...xrffit.pymca_config import pymca_configdict_to_model


class EmissionLineGroup(NamedTuple):
    element: str
    name: str
    counts: Union[float, Sequence]

    @property
    def counts_asarray(self):
        return get_counts_array(self.counts)


class ScatterLineGroup(NamedTuple):
    name: str
    counts: Union[float, Sequence]
    prefix: str = "Scatter"

    @property
    def counts_asarray(self):
        return get_counts_array(self.counts)


def get_counts_array(counts: Union[float, Sequence]):
    if isinstance(counts, Sequence):
        return numpy.asarray(counts)
    return numpy.array((counts))


def xrf_spectra(
    linegroups: Sequence[EmissionLineGroup],
    scattergroups: Sequence[ScatterLineGroup],
    nchannels: int = 1024,
    energy: float = 7.5,
    angle_in: float = 62.0,
    angle_out: float = 49.0,
    flux: float = 1e10,
    elapsed_time: float = 1.0,
    fit_background: bool = False,
    quantification: Optional[str] = None,
    num_layers: int = 1,
) -> Tuple[numpy.ndarray, PyMcaXrfConfiguration]:
    """The returned spectra have the shape `counts.shape + (nchannels,)`"""
    peaks = {linegroup.element: linegroup.name for linegroup in linegroups}
    counts: Dict[str, numpy.ndarray] = {
        f"{group.element} {group.name}": group.counts_asarray for group in linegroups
    }
    counts.update(
        {f"Scatter {group.name}": group.counts_asarray for group in scattergroups}
    )

    # MCA parameters
    zero = 0  # keV
    max_energy = energy + 1.5  # keV
    gain = (max_energy - zero) / nchannels

    # Initialize mca_theory
    mca_theory = McaTheory()

    if quantification == "ref":
        ref_idx = numpy.argmax(
            [linegroup.counts_asarray.mean() for linegroup in linegroups]
        )
        quantification = linegroups[ref_idx].element

    elements = [linegroup.element for linegroup in linegroups]
    configuration = _get_configuration(
        peaks,
        xmin=1,
        xmax=nchannels - 3,
        zero=zero,
        gain=gain,
        energy=energy,
        angle_in=angle_in,
        angle_out=angle_out,
        flux=flux,
        elapsed_time=elapsed_time,
        fit_background=fit_background,
        quantification=quantification,
        elements=elements,
        num_layers=num_layers,
    )
    configuration = pymca_configdict_to_model(
        mca_theory.configure(pymca_configdict_from_model(configuration))
    )

    # Calculate emission lines
    x = numpy.arange(nchannels)
    mca_theory.setData(x=x, y=numpy.zeros(nchannels))
    with _enable_linear_derivatives(mca_theory):
        mca_theory.estimate()

        # Analytical XRF spectrum
        nglobal = mca_theory.NGLOBAL
        if counts:
            shape = next(iter(counts.values())).shape
        else:
            shape = tuple()
        y = numpy.zeros(shape + (nchannels,))
        for i, name in enumerate(mca_theory.PARAMETERS[nglobal:], nglobal):
            normalized_group = mca_theory.linearMcaTheoryDerivative(
                numpy.asarray(mca_theory.parameters), i, x
            )
            assert isinstance(normalized_group, numpy.ndarray)
            if shape:
                y += counts[name][..., None] * normalized_group[None, ...]
            else:
                y += counts[name] * normalized_group
        if fit_background:
            if shape:
                ymin = numpy.clip(numpy.min(y, axis=1), -numpy.inf, 0)
                y += -ymin[:, numpy.newaxis] + 10
            else:
                y += -min(min(y), 0) + 10

    # Configuration for fitting
    mca_theory.config["fit"]["use_limit"] = 1
    mca_theory.config["fit"]["linearfitflag"] = 0

    return y, pymca_configdict_to_model(mca_theory.config)


@contextmanager
def _enable_linear_derivatives(mca_theory: McaTheory):
    linearfitflag = mca_theory.config["fit"]["linearfitflag"]
    mca_theory.config["fit"]["linearfitflag"] = 1
    mca_theory.enableOptimizedLinearFit()
    try:
        yield
    finally:
        mca_theory.disableOptimizedLinearFit()
        mca_theory.config["fit"]["linearfitflag"] = linearfitflag


def _get_configuration(
    peaks: dict,
    xmin: int = 0,
    xmax: int = 0,
    zero: float = 0,
    gain: float = 0.005,
    energy: float = 0,
    angle_in: float = 62.0,
    angle_out: float = 49.0,
    flux: float = 1e10,
    elapsed_time: float = 1.0,
    fit_background: bool = False,
    quantification: Optional[str] = None,
    elements: Optional[List[str]] = None,
    num_layers: int = 1,
) -> PyMcaXrfConfiguration:
    material_config = get_material_config(
        elements=elements,
        num_layers=num_layers,
        angle_in=angle_in,
        angle_out=angle_out,
        quantification=quantification,
    )
    return PyMcaXrfConfiguration(
        peaks=peaks,
        fit=get_fit_config(
            xmin=xmin, xmax=xmax, energy=energy, fit_background=fit_background
        ),
        concentrations=get_quantification_config(
            flux=flux, elapsed_time=elapsed_time, quantification=quantification
        ),
        detector=get_detector(zero=zero, gain=gain),
        peakshape=get_peak_shape(),
        tube=get_tube(),
        **material_config,
    )


def get_quantification_config(
    flux: float = 1e10, elapsed_time: float = 1.0, quantification: Optional[str] = None
) -> dict:
    if quantification:
        if quantification == "fundamental":
            usematrix = 0
            reference = "Auto"
        elif quantification == "auto":
            usematrix = 1
            reference = "Auto"
        else:
            # This is a fix element
            usematrix = 1
            reference = quantification
    else:
        usematrix = 0
        reference = "Auto"
    return {
        "usematrix": usematrix,
        "useattenuators": 1,
        "usemultilayersecondary": 0,
        "usexrfmc": 0,
        "mmolarflag": 0,
        "flux": flux,
        "time": elapsed_time,
        "area": 0.774671,
        "distance": 4.15114,
        "reference": reference,
        "useautotime": 0,
    }


def get_peak_shape():
    return {
        "st_arearatio": 0.0697778,
        "deltast_arearatio": 0.03,
        "fixedst_arearatio": 1,
        "st_sloperatio": 0.0105551,
        "deltast_sloperatio": 0.49,
        "fixedst_sloperatio": 1,
        "lt_arearatio": 0.1,
        "deltalt_arearatio": 0.15,
        "fixedlt_arearatio": 1,
        "lt_sloperatio": 0.1,
        "deltalt_sloperatio": 0.08,
        "fixedlt_sloperatio": 1,
        "step_heightratio": 0.1,
        "deltastep_heightratio": 5e-05,
        "fixedstep_heightratio": 1,
        "eta_factor": 0.02,
        "deltaeta_factor": 0.02,
        "fixedeta_factor": 1,
    }


def get_tube() -> dict:
    return {
        "transmission": 0,
        "voltage": 30.0,
        "anode": "Ag",
        "anodethickness": 0.0002,
        "anodedensity": 10.5,
        "window": "Be",
        "windowthickness": 0.0125,
        "windowdensity": 1.848,
        "filter1": "He",
        "filter1thickness": 0.0,
        "filter1density": 0.000118,
        "alphax": 90.0,
        "alphae": 90.0,
        "deltaplotting": 0.1,
    }


def get_fit_config(
    xmin: int = 0, xmax: int = 0, energy: float = 0, fit_background: bool = False
) -> dict:
    return {
        "deltaonepeak": 0.01,
        "strategy": "SingleLayerStrategy",
        "strategyflag": 0,
        "fitfunction": 0,
        "continuum": 1 if fit_background else 0,
        "fitweight": 1,
        "stripalgorithm": 1,  # SNIP
        "linpolorder": 5,
        "exppolorder": 6,
        "stripconstant": 1.0,
        "snipwidth": 65,
        "stripiterations": 400,
        "stripwidth": 5,
        "stripfilterwidth": 7,
        "stripanchorsflag": 0,
        "maxiter": 50,
        "deltachi": 0.001,
        "xmin": xmin,
        "xmax": xmax,
        "linearfitflag": 1,  # set to 0 after calculating the theoretical spectrum
        "use_limit": 0,  # set to 1 after calculating the theoretical spectrum
        "stripflag": 0 if fit_background else 1,
        "escapeflag": 0,
        "sumflag": 0,
        "scatterflag": 1,
        "hypermetflag": 1,
        "stripanchorslist": [0, 0, 0, 0],
        "energy": [energy],
        "energyweight": [1.0],
        "energyflag": [1],
        "energyscatter": [1],
    }


def get_material_config(
    elements: Optional[List[str]] = None,
    num_layers: int = 1,
    angle_in: float = 62.0,
    angle_out: float = 49.0,
    quantification: Optional[str] = None,
) -> dict:
    if elements is None:
        elements = []
    fixed_internal_reference = quantification not in (None, "fundamental", "auto")
    # WARNING: when quantification=="auto" at least one element in the fit must be in the matrix
    if fixed_internal_reference:
        if quantification not in elements:
            raise ValueError(
                f"'quantification={quantification}' is not in the list of elements {elements}"
            )
        # Put the internal reference in "compound1"
        elements.insert(0, elements.pop(elements.index(quantification)))

    # Split list of elements in layers
    num_elements_per_layer = max(len(elements) // num_layers, 1)
    element_layers = [
        elements[i : i + num_elements_per_layer]
        for i in range(0, len(elements), num_elements_per_layer)
    ]
    num_layers = len(element_layers)

    # Get materials database and define attenuator/filter/detector materials
    materials = get_materials(element_layers)
    attenuators = get_attenuators()

    # Define the matrix
    if num_layers > 1:
        matrix_material = "MULTILAYER"
    else:
        matrix_material = "compound1"
    attenuators["Matrix"] = [
        1,  # enable
        matrix_material,
        2.111,  # density (g/cm^3)
        0.1,  # thickness (cm)
        angle_in,  # incident angle w.r.t sample surface  (deg)
        angle_out,  # outgoing angle w.r.t sample surface (deg)
        0,  # enable scattering angle
        angle_in + angle_out,  # outgoing angle w.r.t primary beam
    ]

    # Define the MULTILAYER composition
    multilayer = {
        f"Layer{i}": [1, f"compound{i}", 2.111, 0.1] for i in range(1, num_layers + 1)
    }

    # Define the iterative fit strategy
    single_layer_strategy = {
        "layer": "Auto",
        "iterations": 3,
        "completer": "-",
        "peaks": "",
        "materials": ["-", "-", "-", "-", "-", "-", "-", "-", "-", "-"],
        "flags": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    }

    return {
        "attenuators": attenuators,
        "materials": materials,
        "multilayer": multilayer,
        "SingleLayerStrategy": single_layer_strategy,
    }


def get_detector(zero: float = 0, gain: float = 0.005) -> dict:
    return {
        "detele": "Si",
        "nthreshold": 4,
        "zero": zero,
        "deltazero": 0.1,
        "fixedzero": 0,  # only free non-linear parameter
        "gain": gain,
        "deltagain": 0.002,
        "fixedgain": 1,
        "noise": -0.0955892,
        "deltanoise": 0.05,
        "fixednoise": 1,
        "fano": 0.0484433,
        "deltafano": 0.114,
        "fixedfano": 1,
        "sum": 0.0,
        "deltasum": 2e-06,
        "fixedsum": 1,
        "ignoreinputcalibration": 0,
    }


def get_attenuators() -> dict:
    return {
        "kapton": [1, "Mylar", 1.4, 0.0004, 1.0],
        "atmosphere": [0, "-", 0.0, 0.0, 1.0],
        "deadlayer": [0, "Si1", 2.33, 0.002, 1.0],
        "absorber": [0, "-", 1.0, 0.1, 1.0],
        "window": [1, "Be", 1.848, 0.0025, 1.0],
        "contact": [0, "moxtekAP3.3", 5.45e-05, 1.0, 1.0],
        "Filter 6": [0, "-", 0.0, 0.0, 1.0],
        "Filter 7": [0, "-", 1.0, 0.1, 1.0],
        "BeamFilter0": [0, "-", 0.0, 0.0, 1.0],
        "BeamFilter1": [0, "-", 0.0, 0.0, 1.0],
        "Detector": [1, "Si1", 2.33, 0.5, 1.0],
    }


def get_materials(element_layers: List[List[str]]) -> dict:
    materials = {**_POLYMERS, **_OTHER_COMPOUNDS}
    for i, elements in enumerate(element_layers, 1):
        num_elements = len(elements)
        materials[f"compound{i}"] = {
            "Comment": f"Compound with elements {' '.join(elements)}",
            "Thickness": 0.0025,
            "Density": 1.42,
            "CompoundFraction": [1.0 / num_elements] * num_elements,
            "CompoundList": [f"{el}1" for el in elements],
        }
    return materials


_POLYMERS = {
    "Mylar": {
        "Comment": "Mylar (Polyethylene Terephthalate) density=1.40 g/cm3",
        "Density": 1.4,
        "CompoundFraction": [0.041959, 0.625017, 0.333025],
        "CompoundList": ["H1", "C1", "O1"],
    },
    "Kapton": {
        "Comment": "Kapton 100 HN 25 micron density=1.42 g/cm3",
        "Thickness": 0.0025,
        "Density": 1.42,
        "CompoundFraction": [0.628772, 0.066659, 0.304569],
        "CompoundList": ["C1", "N1", "O1"],
    },
    "Teflon": {
        "Comment": "Teflon density=2.2 g/cm3",
        "Density": 2.2,
        "CompoundFraction": [0.240183, 0.759817],
        "CompoundList": ["C1", "F1"],
    },
    "Viton": {
        "Comment": "Viton Fluoroelastomer density=1.8 g/cm3",
        "Density": 1.8,
        "CompoundFraction": [0.009417, 0.280555, 0.710028],
        "CompoundList": ["H1", "C1", "F1"],
    },
    "moxtekAP3.3": {
        "Comment": "New Material",
        "Thickness": 1.0,
        "Density": 5.45e-05,
        "CompoundFraction": [0.06, 0.62, 0.055, 0.155, 0.106],
        "CompoundList": ["B", "C", "N", "O", "Al"],
    },
}


_OTHER_COMPOUNDS = {
    "Air": {
        "Comment": "Dry Air (Near sea level) density=0.001204790 g/cm3",
        "Thickness": 1.0,
        "Density": 0.0012048,
        "CompoundFraction": [0.000124, 0.75527, 0.23178, 0.012827, 3.2e-06],
        "CompoundList": ["C1", "N1", "O1", "Ar1", "Kr1"],
    },
    "Water": {
        "Comment": "Water density=1.0 g/cm3",
        "CompoundFraction": 1.0,
        "Thickness": 1.0,
        "Density": 1.0,
        "CompoundList": "H2O1",
    },
    "SingleLayerStrategyMaterial": {
        "Comment": "Last Single Layer Strategy iteration",
        "Thickness": 0.0025,
        "Density": 1.42,
        "CompoundFraction": [],
        "CompoundList": [],
    },
}



from typing import Tuple

import numpy
from PyMca5.PyMcaPhysics.xrf import ClassMcaTheory


def channel_info(
    mca_theory: ClassMcaTheory,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """Return MCA channel indices use for fitting and the raw data."""
    channels = mca_theory.xdata[:, 0]
    raw_channels = mca_theory.xdata0
    return channels, raw_channels


def energy_info(mca_theory: ClassMcaTheory, channels: numpy.ndarray) -> numpy.ndarray:
    """Return MCA bin energies use for fitting and the raw data."""
    zero, gain = mca_theory.parameters[:2]
    energies = zero + gain * channels
    return energies



from ewoksorange.gui.owwidgets.meta import ow_build_opts
from ewoksorange.gui.owwidgets.threaded import OWEwoksWidgetOneThread
from ewoksorange.gui.widgets.parameter_form import ParameterForm

from ewoksfluo.gui.data_viewer import DataViewer
from ewoksfluo.io.hdf5 import split_h5uri
from ewoksfluo.tasks.sum_detectors import sum_results


class OWSumXrfResults(
    OWEwoksWidgetOneThread, **ow_build_opts, ewokstaskclass=sum_results.SumXrfResults
):
    name = "Sum Fit Results"
    description = "Add single-scan XRF results of multiple detectors"

    def __init__(self):
        super().__init__()
        self._init_control_area()
        self._init_main_area()

    def _init_control_area(self):
        super()._init_control_area()
        self._default_inputs_form = ParameterForm(parent=self.controlArea)
        defaults = {"output_root_uri": "processed_data.h5"}
        values = self.get_default_input_values(include_missing=True, defaults=defaults)
        self.update_default_inputs(**values)

        parameters = {
            "bliss_scan_uri": {
                "label": "URI scan",
                "value_for_type": "",
                "select": "h5group",
            },
            "xrf_results_uris": {
                "label": "URIs to fit results of each detector",
                "value_for_type": "",
            },
            "detector_normalization_template": {
                "label": "Detector weight formula",
                "value_for_type": "",
            },
            "output_root_uri": {
                "label": "HDF5 output URI",
                "value_for_type": "",
                "select": "h5group",
            },
        }

        for name, kw in parameters.items():
            self._default_inputs_form.addParameter(
                name,
                value=values[name],
                value_change_callback=self._default_inputs_changed,
                **kw,
            )

    def _default_inputs_changed(self):
        self.update_default_inputs(**self._default_inputs_form.get_parameter_values())
        self._update_input_data()
        self._update_output_data()

    def handleNewSignals(self) -> None:
        self._update_input_data()
        super().handleNewSignals()

    def task_output_changed(self):
        self._update_output_data()

    def _init_main_area(self):
        super()._init_main_area()
        layout = self._get_main_layout()

        self._viewer = DataViewer(parent=self.mainArea)
        self._viewer.setVisible(True)
        layout.addWidget(self._viewer)
        layout.setStretchFactor(self._viewer, 1)

        self._update_output_data()

    def _update_input_data(self):
        dynamic = self.get_dynamic_input_names()
        for name in self.get_input_names():
            self._default_inputs_form.set_parameter_enabled(name, name not in dynamic)

    def _update_output_data(self):
        self._refresh_output_file()

    def _refresh_output_file(self):
        uri = self.get_task_input_value("output_root_uri")
        if uri:
            self._viewer.updateFile(split_h5uri(uri)[0])



from pint import Unit
from pint import UnitRegistry

_UNIT_REGISTRY = None


def unit_registry():
    global _UNIT_REGISTRY
    if _UNIT_REGISTRY is None:
        _UNIT_REGISTRY = UnitRegistry()
    return _UNIT_REGISTRY


def units_as_str(units: Unit) -> str:
    return f"{units:~}"



from typing import Optional

import numpy
from PyMca5.PyMcaPhysics.xrf.LegacyMcaAdvancedFitBatch import (
    McaAdvancedFitBatch as LegacyMcaAdvancedFitBatch,
)
from PyMca5.PyMcaPhysics.xrf.McaAdvancedFitBatch import McaAdvancedFitBatch

from ..pymca_config import PyMcaXrfConfiguration
from ..pymca_config import pymca_configdict_from_model
from ._utils.parse_native import parse_native_fit_output
from ._utils.parse_native import parse_native_legacy_fit_output
from .types import XRFBatchFitResult


def fit_xrf_spectra(
    xrf_spectra: numpy.ndarray,
    configuration: PyMcaXrfConfiguration,
    quantification: Optional[bool] = None,
    individual_weights: Optional[bool] = None,
    live_times: Optional[numpy.ndarray] = None,
    positive_peak_areas: Optional[bool] = None,
    diagnostics: Optional[bool] = None,
) -> XRFBatchFitResult:
    """Slow fitting refers to fitting XRF spectra one-by-one.
    This function uses the high-level API from PyMca.

    :param xrf_spectra: shape `(num_spectra, num_channels)`
    :param configuration: PyMca configuration.
    :param quantification: Calculate mass fractions from peak area's.
    :param individual_weights: When fitting with weights, use the weight of each
                                spectrum (slow) instead of the average weight.
    :param live_times: 1D array with one value for each XRF spectrum.
                       Only used for mass fractions with fundamental parameters.
    :param positive_peak_areas:
    :param diagnostics: fit model and residuals.
    """

    if quantification is None:
        quantification = True
    if individual_weights is None:
        individual_weights = True
    if positive_peak_areas is None:
        positive_peak_areas = True
    if diagnostics is None:
        diagnostics = False

    if not positive_peak_areas:
        raise NotImplementedError("'positive_peak_areas' must be True")
    if live_times is not None:
        raise NotImplementedError("'live_times' is not supported")
    if not individual_weights:
        raise NotImplementedError("'individual_weights' must be True")

    batch = McaAdvancedFitBatch(
        None,
        filelist=xrf_spectra,
        concentrations=int(quantification),
        fitfiles=0,
        fitconcfile=0,
        nosave=True,
        saveResiduals=True,
        saveFit=True,
        saveData=True,
        diagnostics=True,
        saveFOM=True,
    )

    # WARNING: McaAdvancedFitBatch ignores `use_limit=False`
    #          and always uses xmin/xmax from the configuration.
    if not configuration.fit.use_limit:
        configuration.fit.xmin = 0
        configuration.fit.xmax = xrf_spectra.shape[1]

    _ = batch.mcafit.configure(pymca_configdict_from_model(configuration))

    with numpy.errstate(over="ignore"):  # Stores results in float32
        batch.processList()

    return parse_native_fit_output(
        xrf_spectra,
        batch.mcafit,
        batch.outbuffer,
        quantification,
        diagnostics,
        fast_fitting=False,
    )


class _LegacyMcaAdvancedFitBatch(LegacyMcaAdvancedFitBatch):
    def __log(self, text):
        # Base class does `print`
        pass


def fit_xrf_spectra_legacy(
    xrf_spectra: numpy.ndarray,
    configuration: PyMcaXrfConfiguration,
    quantification: Optional[bool] = None,
    individual_weights: Optional[bool] = None,
    live_times: Optional[numpy.ndarray] = None,
    positive_peak_areas: Optional[bool] = None,
    diagnostics: Optional[bool] = None,
) -> XRFBatchFitResult:
    """Slow fitting refers to fitting XRF spectra one-by-one.
    This function uses the high-level API from PyMca.

    :param xrf_spectra: shape `(num_spectra, num_channels)`
    :param configuration: PyMca configuration.
    :param quantification: Calculate mass fractions from peak area's.
    :param individual_weights: When fitting with weights, use the weight of each
                                spectrum (slow) instead of the average weight.
    :param live_times: 1D array with one value for each XRF spectrum.
                       Only used for mass fractions with fundamental parameters.
    :param positive_peak_areas:
    :param diagnostics: fit model and residuals.
    """

    if quantification is None:
        quantification = True
    if individual_weights is None:
        individual_weights = True
    if positive_peak_areas is None:
        positive_peak_areas = True
    if diagnostics is None:
        diagnostics = False

    if not positive_peak_areas:
        raise NotImplementedError("'positive_peak_areas' must be True")
    if live_times is not None:
        raise NotImplementedError("'live_times' is not supported")
    if not individual_weights:
        raise NotImplementedError("'individual_weights' must be True")

    batch = _LegacyMcaAdvancedFitBatch(
        None,
        filelist=xrf_spectra,
        concentrations=int(quantification),
        fitfiles=0,
        nosave=True,
    )

    class Dummy:
        def getConcentrationsAsAscii(self, _):
            # Prevent saving of _concentrations.txt file
            return []

    batch._toolConversion = Dummy()

    # WARNING: McaAdvancedFitBatch ignores `use_limit=False`
    #          and always uses xmin/xmax from the configuration.
    if not configuration.fit.use_limit:
        configuration.fit.xmin = 0
        configuration.fit.xmax = xrf_spectra.shape[1]

    _ = batch.mcafit.configure(pymca_configdict_from_model(configuration))

    with numpy.errstate(over="ignore"):  # Stores results in float32
        batch.processList()

    images = batch._McaAdvancedFitBatch__images
    sigmas = batch._McaAdvancedFitBatch__sigmas

    parameter_names = list(sigmas)
    parameters = numpy.array([images[name] for name in parameter_names])
    uncertainties = numpy.array([sigmas[name] for name in parameter_names])
    outputdict = {"parameters": parameters, "uncertainties": uncertainties}

    names = list(parameter_names)
    if quantification:
        concentrations = []
        for name in parameter_names:
            key = f"{name} mass fraction"
            if key in images:
                concentrations.append(images[key])
                names.append(f"C({name})")

        layeri = 0
        has_layeri = True
        while has_layeri:
            layeri += 1
            has_layeri = False
            for name in parameter_names:
                key = f"{name} Layer{layeri}"
                if key in images:
                    concentrations.append(images[key])
                    names.append(f"C({name})-Layer{layeri}")
                    has_layeri = True

        outputdict["concentrations"] = numpy.array(concentrations)

    outputdict["names"] = names

    return parse_native_legacy_fit_output(
        xrf_spectra,
        batch.mcafit,
        outputdict,
        quantification,
        diagnostics,
        fast_fitting=False,
    )



import logging

from ..model import PyMcaXrfConfiguration

logger = logging.getLogger(__name__)


def ensure_matrix(configuration: PyMcaXrfConfiguration) -> None:
    attenuators = configuration.attenuator_views
    matrix = attenuators["Matrix"]

    if matrix.material.upper() == "MULTILAYER":
        multilayer = configuration.multilayer_views
        matrix_defined = False
        for layer in multilayer.values():
            if layer.use and layer.material:
                matrix_defined = True
                break
    else:
        matrix_defined = bool(matrix.material)

    if matrix.enabled and matrix_defined:
        return

    matrix.enabled = 1
    matrix.material = "H1"
    matrix.density = 1.0
    matrix.thickness = 0.0
    logger.warning(
        "PyMca configuration: matrix is required but not defined. Use dummy matrix with composition 'H1', density 1.0 g/cm^3 and thickness 0.0 cm."
    )



import datetime
from typing import Optional

import numpy

from ..handlers import AbstractOutputHandler
from .abstract import AbstractOutputBuffer


class ExternalOutputBuffer(AbstractOutputBuffer):
    """This is an output buffer of an external output handler."""

    def __init__(
        self,
        output_handler: AbstractOutputHandler,
        diagnostics: bool = False,
        figuresofmerit: bool = False,
    ):
        self._output_handler = output_handler
        super().__init__(
            saveResiduals=diagnostics,
            saveFit=diagnostics,
            saveData=diagnostics,
            diagnostics=diagnostics,
            saveFOM=figuresofmerit,
            nosave=True,
        )

    def _allocateRam(self, label, data=None, **kwargs) -> numpy.ndarray:
        # Fix bug in PyMca: https://github.com/silx-kit/pymca/issues/1123
        ids = set(map(id, self._buffers.values()))
        if id(data) in ids:
            data = data.copy()
        return super()._allocateRam(label, data=data, **kwargs)

    @property
    def xrf_results_uri(self) -> Optional[str]:
        return self._output_handler.xrf_results_uri

    @property
    def already_existed(self) -> bool:
        return self._output_handler.already_existed

    @property
    def output_root_uri(self) -> Optional[str]:
        return self._output_handler.output_root_uri

    def save(self):
        self._create_configuration()
        for group_name in self._results:
            self._create_nxdata_group(group_name)

    def _create_configuration(self):
        configuration = self._info["configuration"]
        data = {
            "@NX_class": "NXnote",
            "type": "text/plain",
            "data": configuration.tostring(),
            "date": datetime.datetime.now().astimezone().isoformat(),
        }
        self._output_handler.create_group("configuration", data=data)

    def _create_nxdata_group(self, group_name: str):
        group_data = self._results[group_name]

        data = {"@NX_class": "NXdata"}

        if group_data["axesused"]:
            last_axis_name = group_data["axesused"][-1]
            for axis_name, axis_data, axis_attrs in group_data["axes"]:
                if axis_name != last_axis_name:
                    continue
                data["@axes"] = [axis_name]
                data[axis_name] = axis_data
                for attr_name, attr_value in axis_attrs.items():
                    data[f"{axis_name}@{attr_name}"] = attr_value
                break

        self._output_handler.create_group(group_name, data=data)

        for signal_name, signal_data, _ in group_data["_signals"]:
            data = signal_data["data"]
            if data.ndim > 1:
                # Flatten the first two scan dimensions in (nscandim1, nscandim2, ...).
                data = data.reshape(data.shape[0] * data.shape[1], *data.shape[2:])

            nxdata_handler = self._output_handler.create_nxdata_handler(
                group_name, signal_name, len(data)
            )
            nxdata_handler.add_points(data)



import logging
import re
from typing import Any
from typing import Tuple

from ..io.hdf5 import FileReadAccess
from ..io.hdf5 import split_h5uri
from ..math.expression import eval_expression
from ..math.expression import expression_variables

logger = logging.getLogger(__name__)


def format_expression_template(template: str, name: str) -> str:
    n = len(re.findall(r"\{\}", template))
    return template.format(*[name] * n)


def eval_hdf5_expression(
    data_uri: str, expression: str, start_var: str = "<", end_var: str = ">"
) -> Any:
    """Evaluate an arithmetic expression with python and numpy arithmetic
    on HDF5 datasets.

    :param data_uri: HDF5 root URI
    :param expression: arithmetic expression where datasets are define as
                       :code:`"<subgroup/data>"` where :code:`"subgroup/data"`
                       is relative to :code:`data_uri`.
    :param start_var: marks the start of a variable name
    :param end_var: marks the end of a variable name
    """
    logger.debug("Evaluate expression %r for %r", expression, data_uri)

    data_file, data_h5path = split_h5uri(data_uri)

    with FileReadAccess(data_file) as h5file:
        if not data_h5path:
            data_h5path = "/"
        data_root = h5file[data_h5path]

        def get_data(path: str) -> Tuple[str, Any]:
            if path not in data_root:
                raise RuntimeError(
                    f"HDF5 dataset '{data_file}::{data_root.name}/{path}' does not exist"
                )
            return path, data_root[path][()]

        final_expression, variables, _ = expression_variables(
            expression, get_data, start_var=start_var, end_var=end_var
        )

    return eval_expression(final_expression, variables)



from PyMca5.PyMcaGui.pymca.RGBCorrelator import RGBCorrelator


class XrfResultViewer(RGBCorrelator):
    def removeAllImages(self):
        for label in list(self._imageList):
            self.removeImage(label)



from ewoksfluo.gui.mesh_widget import OWMeshWidget
from ewoksfluo.tasks.example_data.tasks import MosaicMeshSingleDetector


class OWMosaicMeshSingleDetector(OWMeshWidget, ewokstaskclass=MosaicMeshSingleDetector):
    name = "Mosaic Mesh (single detector)"
    description = "XRF test data of one scan with one detector"

    def _init_control_area(self):
        super()._init_control_area(stack=False, multidetector=False, mosaic=True)



from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Sequence
from typing import Union

from ewokscore import Task
from ewokscore.model import BaseOutputModel
from pydantic import Field

from ...io import hdf5
from ...math.regular_grid import ScatterDataInterpolator
from ...math.rounding import round_to_significant
from .. import nexus_utils
from ..models import FluoInputs
from ..xrf_results import get_xrf_result_groups
from . import positioners_utils


class Inputs(FluoInputs):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )

    output_root_uri: str = Field(
        description="Output root URI.",
        examples=[
            "/results/dataset.h5",
            "/results/dataset.h5::/1.1",
            "/results/dataset.h5::/1.1/regrid",
        ],
    )

    xrf_results_uri: Optional[str] = Field(
        description="Previous XRF NXprocess URI to regrid.",
        examples=["/results/dataset.h5::/1.1/norm/results"],
    )

    positioners: Union[List[str], str, None] = Field(default=None)
    positioner_uri_template: str = Field(default="measurement/{}")
    ignore_positioners: Optional[List[str]] = Field(default=None)
    interpolate: Literal["none", "nearest", "linear", "cubic"] = Field(
        default="nearest"
    )
    resolution: Optional[dict] = Field(default=None)

    axes_units: Dict[str, str] = Field(
        default_factory=dict,
        description="Axes units to be used when missing.",
        examples=[{"nsy": "mm", "nsz": "mm", "nspy": "um", "nspz": "um"}],
    )


class Outputs(BaseOutputModel):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )
    output_root_uri: str = Field(
        description="Output NXentry URI.", examples=["/results/dataset.h5::/1.1"]
    )
    xrf_results_uri: str = Field(
        description="Output NXprocess results URI.",
        examples=["/results/dataset.h5::/1.1/regrid/results"],
    )


class RegridXrfResults(Task, input_model=Inputs, output_model=Outputs):
    """Regrid single-scan XRF results on a regular grid by interpolation."""

    def run(self):
        start_time = nexus_utils.now()
        output_root_uri = self.inputs.output_root_uri

        with nexus_utils.save_in_ewoks_subprocess(
            output_root_uri, start_time, {}, default_levels=("results", "regrid")
        ) as (regrid_results, already_existed):
            if not already_existed:
                self._regrid(regrid_results)

            self.outputs.xrf_results_uri = (
                f"{regrid_results.file.filename}::{regrid_results.name}"
            )
        self.outputs.bliss_scan_uri = self.inputs.bliss_scan_uri
        self.outputs.output_root_uri = output_root_uri

    def _regrid(self, regrid_results):
        position_suburis = self._get_position_suburis(
            self.inputs.bliss_scan_uri,
            ignore_positioners=self.inputs.ignore_positioners,
        )

        coordinates, grid_names, grid_units = positioners_utils.read_position_suburis(
            self.inputs.bliss_scan_uri,
            position_suburis,
            axes_units=self.inputs.axes_units,
        )

        resolution = self.inputs.resolution
        if resolution:
            resolution = [resolution[name] for name in grid_names]
        else:
            resolution = None

        interpolator = ScatterDataInterpolator(
            coordinates,
            grid_names,
            grid_units,
            method=self.inputs.interpolate,
            resolution=resolution,
        )

        xrf_results_filename, xrf_results_h5path = hdf5.split_h5uri(
            self.inputs.xrf_results_uri
        )
        with hdf5.FileReadAccess(xrf_results_filename) as xrf_results_file:
            xrf_results_grp = xrf_results_file[xrf_results_h5path]
            if not hdf5.is_group(xrf_results_grp):
                raise TypeError(f"'{self.inputs.xrf_results_uri}' must be a group")

            nxdata_groups = get_xrf_result_groups(xrf_results_grp)

            title_parts = list()
            for group_name in reversed(list(nxdata_groups)):
                input_grp = xrf_results_grp[group_name]
                input_datasets = {
                    dset_name: dset
                    for dset_name, dset in input_grp.items()
                    if hdf5.is_dataset(dset) and dset_name not in grid_names
                }
                if not input_datasets:
                    # NXdata group which does not plot scan data
                    continue

                # NXdata signals
                output_grp = nexus_utils.create_nxdata(regrid_results, group_name)
                for dset_name, dset in input_datasets.items():
                    output_grp.create_dataset(
                        dset_name, data=interpolator.regrid(dset[()])
                    )

                nexus_utils.set_nxdata_signals(
                    output_grp, signals=tuple(input_datasets.keys())
                )

                # NXdata axes
                axes = list()
                for i, (axisname, arr) in enumerate(
                    zip(grid_names, interpolator.grid_axes)
                ):
                    axes.append(axisname)
                    dset = output_grp.create_dataset(axisname, data=arr)
                    if interpolator.units is not None:
                        dset.attrs["units"] = interpolator.units
                        dset.attrs["long_name"] = f"{axisname} ({interpolator.units})"
                    output_grp.create_dataset(f"{axisname}_indices", data=i)
                    title_parts.append(
                        (
                            axisname,
                            len(arr),
                            round_to_significant(abs(arr[1] - arr[0])),
                            interpolator.units,
                        )
                    )
                output_grp.attrs["axes"] = axes

            interpolator.save_coordinates_as_nxdata(regrid_results)

            if title_parts:
                title = [
                    f"{axisname} ({size} x {resolution} {units})"
                    for axisname, size, resolution, units in title_parts
                ]
                regrid_results["title"] = " x ".join(title)

    def _get_position_suburis(
        self, bliss_scan_uri: str, ignore_positioners: Optional[Sequence[str]] = None
    ) -> List[str]:
        positioners = self.inputs.positioners
        if not positioners:
            return positioners_utils.get_scan_position_suburis(
                bliss_scan_uri, ignore_positioners=ignore_positioners
            )
        if isinstance(positioners, str):
            positioners = [positioners]
        template = self.inputs.positioner_uri_template
        return [template.format(s) for s in positioners]



from typing import Dict
from typing import List

from ewokscore import execute_graph
from ewoksorange.gui.workflows.owscheme import ows_to_ewoks

try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files


def test_multi_scan_multi_detector_without_qt(tmp_path):
    from orangecontrib.ewoksfluo.categories.demo import tutorials

    filename = resource_files(tutorials).joinpath("mesh_stack_multi_detector.ows")
    assert_multi_scan_multi_detector_without_qt(filename, tmp_path)


def test_multi_scan_multi_detector_with_qt(ewoks_orange_canvas, tmp_path):
    from orangecontrib.ewoksfluo.categories.demo import tutorials

    filename = resource_files(tutorials).joinpath("mesh_stack_multi_detector.ows")
    assert_multi_scan_multi_detector_with_qt(ewoks_orange_canvas, filename, tmp_path)


def assert_multi_scan_multi_detector_without_qt(filename, tmp_path):
    """Execute workflow after converting it to an ewoks workflow"""
    graph = ows_to_ewoks(filename)
    outputs = execute_graph(
        graph, inputs=get_inputs(tmp_path), outputs=[{"all": True}], merge_outputs=False
    )
    expected = get_expected_outputs(tmp_path)
    label_to_id = {
        attrs["label"]: node_id for node_id, attrs in graph.graph.nodes.items()
    }
    expected = {label_to_id[k]: v for k, v in expected.items()}
    assert outputs == expected


def assert_multi_scan_multi_detector_with_qt(ewoks_orange_canvas, filename, tmp_path):
    """Execute workflow using the Qt widgets and signals"""
    ewoks_orange_canvas.load_graph(str(filename), inputs=get_inputs(tmp_path))
    ewoks_orange_canvas.start_workflow()
    ewoks_orange_canvas.wait_widgets(timeout=30)
    outputs = dict(ewoks_orange_canvas.iter_output_values())
    assert outputs == get_expected_outputs(tmp_path)


def get_inputs(tmp_path) -> List[dict]:
    return [
        {
            "label": "Mesh stack (multi detector)",
            "name": "output_filename",
            "value": str(tmp_path / "input.h5"),
        },
        {
            "label": "Fit stack (multi detector)",
            "name": "output_root_uri",
            "value": str(tmp_path / "output.h5::/1.1"),
        },
    ]


def get_expected_outputs(tmp_path) -> Dict[str, dict]:
    bliss_scan_uris = [str(tmp_path / f"input.h5::/{i}.1") for i in range(1, 4)]
    output_root_uri = str(tmp_path / "output.h5::/1.1")
    return {
        "Mesh stack (multi detector)": {
            "config": str(tmp_path / "input.h5::/1.1/theory/configuration/data"),
            "configs": [str(tmp_path / "input.h5::/1.1/theory/configuration/data")] * 2,
            "detector_names": ["mca0", "mca1"],
            "expo_time": 0.1,
            "filenames": [str(tmp_path / "input.h5")],
            "monitor_name": "I0",
            "monitor_normalization_template": "1000000/<instrument/{}/data>",
            "detector_normalization_template": "0.1/<instrument/{}/live_time>",
            "scan_ranges": [(1, 3)],
        },
        "Pick scans": {"bliss_scan_uris": bliss_scan_uris},
        "Fit stack (multi detector)": {
            "xrf_results_uris": [
                str(tmp_path / "output.h5::/1.1/fit/mca0/results"),
                str(tmp_path / "output.h5::/1.1/fit/mca1/results"),
            ],
            "bliss_scan_uris": bliss_scan_uris,
            "detector_names": ["mca0", "mca1"],
            "output_root_uri": output_root_uri,
        },
        "Sum Fit Results Stack": {
            "xrf_results_uri": str(tmp_path / "output.h5::/1.1/sum/results"),
            "bliss_scan_uris": bliss_scan_uris,
            "output_root_uri": output_root_uri,
        },
        "Normalize Stack": {
            "xrf_results_uri": str(tmp_path / "output.h5::/1.1/norm/results"),
            "bliss_scan_uris": bliss_scan_uris,
            "output_root_uri": output_root_uri,
        },
        "Raw Counters Stack": {
            "xrf_results_uri": str(tmp_path / "output.h5::/1.1/merge/results"),
            "bliss_scan_uris": bliss_scan_uris,
            "output_root_uri": output_root_uri,
        },
        "Regrid Stack": {
            "xrf_results_uri": str(tmp_path / "output.h5::/1.1/regrid/results"),
            "bliss_scan_uris": bliss_scan_uris,
            "output_root_uri": output_root_uri,
        },
    }



import numbers
from typing import Any
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Tuple

import h5py
import numpy


def pad_length(*all_arrays: Tuple[Dict[str, Any]]) -> Optional[int]:
    """Return the length to which all arrays should be padded to be the same size.
    Returns None if all arrays already have the same size or no arrays are found.
    """
    lengths = set()

    for arrays in all_arrays:
        for value in arrays.values():
            if _is_number_array(value):
                lengths.add(len(value))

    if len(lengths) > 1:
        return max(lengths)
    else:
        return None


def pad_arrays(max_len: int, *all_arrays: Tuple[Dict[str, Any]]) -> None:
    """Pad all sequence-like values in all dictionaries to the specified max_len with nan."""
    for arrays in all_arrays:
        for key, value in list(arrays.items()):
            if _is_number_array(value):
                arrays[key] = pad_array(value, max_len)


def pad_array(value: Sequence, max_len: int) -> Sequence:
    """Pad array when the length is smaller than max_len."""
    nextra = max_len - len(value)
    if nextra <= 0:
        return value
    arr = numpy.asarray(value, dtype=float)
    pad_width = [(0, nextra)] + [(0, 0)] * (arr.ndim - 1)
    return numpy.pad(arr, pad_width, mode="constant", constant_values=numpy.nan)


def _is_number_array(value: Any) -> bool:
    if isinstance(value, (h5py.Dataset, numpy.ndarray)):
        return numpy.issubdtype(value.dtype, numpy.number)
    if isinstance(value, Sequence) and value:
        return isinstance(value[0], numbers.Number)
    return False



import h5py
import numpy
import pytest

from ..tasks.sum_detectors import sum_utils


@pytest.mark.parametrize("array_chunk_size", [1, 2, 3])
@pytest.mark.parametrize("normalize", [True, False])
def test_sum_spectra(tmp_path, monkeypatch, array_chunk_size, normalize):
    filename = str(tmp_path / "data.h5")

    npts = 10
    data = {
        "mca_det0": {
            "data": numpy.random.uniform(10, 100, (npts, 2048)),
            "live_time": numpy.random.uniform(0.05, 0.099, npts),
        },
        "mca_det1": {
            "data": numpy.random.uniform(10, 100, (npts, 2048)),
            "live_time": numpy.random.uniform(0.05, 0.099, npts),
        },
    }

    if normalize:
        mca_det0_weighted = (
            data["mca_det0"]["data"]
            / data["mca_det0"]["live_time"][:, numpy.newaxis]
            * 0.1
        )
        mca_det1_weighted = (
            data["mca_det1"]["data"]
            / data["mca_det1"]["live_time"][:, numpy.newaxis]
            * 0.1
        )
        expected_sum_spectra = mca_det0_weighted + mca_det1_weighted
    else:
        expected_sum_spectra = data["mca_det0"]["data"] + data["mca_det1"]["data"]

    with h5py.File(filename, "w") as f:
        f["/1.1/instrument/mca_det0/data"] = data["mca_det0"]["data"]
        f["/1.1/instrument/mca_det0/live_time"] = data["mca_det0"]["live_time"]
        f["/1.1/instrument/mca_det1/data"] = data["mca_det1"]["data"]
        f["/1.1/instrument/mca_det1/live_time"] = data["mca_det1"]["live_time"]

    def patched_array_chunk_size(*args, **_):
        return array_chunk_size

    monkeypatch.setattr(sum_utils, "array_chunk_size", patched_array_chunk_size)

    if normalize:
        detector_normalization_template = "0.1/<instrument/{}/live_time>"
    else:
        detector_normalization_template = None

    sum_spectra = sum_utils.sum_spectra_from_hdf5(
        bliss_scan_uri=f"{filename}::/1.1",
        xrf_spectra_uri_template="instrument/{}/data",
        detector_normalization_template=detector_normalization_template,
        detector_names=["mca_det0", "mca_det1"],
    )

    numpy.testing.assert_allclose(sum_spectra, expected_sum_spectra)



from abc import abstractmethod
from typing import Optional

from PyMca5.PyMcaPhysics.xrf.XRFBatchFitOutput import OutputBuffer as _OutputBuffer


class AbstractOutputBuffer(_OutputBuffer):

    @property
    @abstractmethod
    def xrf_results_uri(self) -> Optional[str]:
        pass

    @property
    @abstractmethod
    def already_existed(self) -> bool:
        pass



"""Read data from ZAP scans in SPEC"""

import os
import re
from glob import glob
from typing import Dict
from typing import Iterator
from typing import Sequence
from typing import Tuple

import fabio
import numpy

from .types import ScanData
from .types import ZapInfo

_REMOVE_COUNTER_PREFIXES = {"zap_p201_": "", "zap_": "", "arr_": "", "xmap_": ""}

_REMOVE_EDF_HEADER_KEYS = (
    "EDF_BinarySize",
    "EDF_HeaderSize",
    "Dim_1",
    "Dim_2",
    "Image",
    "Size",
    "xdata",
    "xnb",
    "acquisition_time",
)


_XMAP_COUNTER_PATTERN = re.compile("^xmap_(.+?)(Sum|[0-9]+)$")


def rename_counter(name: str) -> str:
    for substr, substrnew in _REMOVE_COUNTER_PREFIXES.items():
        name = name.replace(substr, substrnew)
    return name


def iter_zap_data(
    zap_info: ZapInfo, exclude_mca_detectors: Sequence[int], motor_names: Sequence[str]
) -> Iterator[ScanData]:
    """Yields MCA spectra, MCA statistics and counters from EDF files"""
    yield from _iter_xia_ctr_data(zap_info, exclude_mca_detectors, motor_names)
    static_data = dict()
    yield from _iter_xia_stats_data(zap_info, exclude_mca_detectors, static_data)
    for detnr in range(static_data.get("ndet", 0)):
        if detnr in exclude_mca_detectors:
            continue
        yield from _iter_xia_mca_data(zap_info, detnr)
    yield from _iter_lima_data(zap_info)


def get_zap_positioners(zap_info: ZapInfo) -> Dict[str, float]:
    pattern = os.path.join(
        zap_info.directory,
        f"{zap_info.radix}_xiast_{zap_info.scannb:04d}_{zap_info.zapnb:04d}_[0-9][0-9][0-9][0-9].edf",
    )
    files = sorted(glob(pattern))
    if not files:
        return dict()
    with fabio.open(files[0]) as f:
        return _get_motors_from_xia_header(f.header)


def _iter_xia_stats_data(
    zap_info: ZapInfo, exclude_mca_detectors: Sequence[int], static_data: dict
) -> Iterator[ScanData]:
    """We have one xiast file for each line in a 2D scan. The 2D data in
    each xiast file has the shape `(ncols, 6*n)` where

        * `ncols` the number of points in each line of the 2D scan
        * `n`: the number of detectors
        * `6`: the number of statistics recorded per detector

    The columns of the 2D data in each xiast file are:

    .. code-block::

      |DETNR1|EVENTS1|ICR1|OCR1|TLT1|DT1|...|DETNRn|EVENTSn|ICRn|OCRn|TLTn|DTn

    """
    pattern = os.path.join(
        zap_info.directory,
        f"{zap_info.radix}_xiast_{zap_info.scannb:04d}_{zap_info.zapnb:04d}_[0-9][0-9][0-9][0-9].edf",
    )
    files = sorted(glob(pattern))
    if not files:
        return 0
    for filename in files:
        nstats = 6
        with fabio.open(filename) as f:
            data = f.data
            ncols, ntotalstats = data.shape
            ndet = ntotalstats // nstats
            if not static_data:
                static_data["ndet"] = ndet

            (
                detnr,
                events,
                trigger_count_rate,
                event_count_rate,
                trigger_live_time,
                fractional_dead_time,
            ) = data.reshape((ncols, ndet, nstats)).T

            trigger_live_time = trigger_live_time / 1000
            fractional_dead_time = fractional_dead_time / 100
            triggers = trigger_count_rate * trigger_live_time
            elapsed_time = events / event_count_rate
            live_time = events / trigger_count_rate

            for i in range(ndet):
                j = detnr[i][0]
                if j in exclude_mca_detectors:
                    continue
                group = f"xia_det{j}"
                yield ScanData(
                    group=group,
                    name="events",
                    detector_type="mca",
                    data=events[i],
                    local_alias="",
                    global_alias=f"{group}_events",
                )
                yield ScanData(
                    group=group,
                    name="trigger_count_rate",
                    detector_type="mca",
                    data=trigger_count_rate[i],
                    local_alias="",
                    global_alias=f"{group}_trigger_count_rate",
                )
                yield ScanData(
                    group=group,
                    name="event_count_rate",
                    detector_type="mca",
                    data=event_count_rate[i],
                    local_alias="",
                    global_alias=f"{group}_event_count_rate",
                )
                yield ScanData(
                    group=group,
                    name="trigger_live_time",
                    detector_type="mca",
                    data=trigger_live_time[i],
                    local_alias="",
                    global_alias=f"{group}_trigger_live_time",
                )
                yield ScanData(
                    group=group,
                    name="fractional_dead_time",
                    detector_type="mca",
                    data=fractional_dead_time[i],
                    local_alias="",
                    global_alias=f"{group}_fractional_dead_time",
                )
                yield ScanData(
                    group=group,
                    name="triggers",
                    detector_type="mca",
                    data=triggers[i],
                    local_alias="",
                    global_alias=f"{group}_triggers",
                )
                yield ScanData(
                    group=group,
                    name="elapsed_time",
                    detector_type="mca",
                    data=elapsed_time[i],
                    local_alias="",
                    global_alias=f"{group}_elapsed_time",
                )
                yield ScanData(
                    group=group,
                    name="live_time",
                    detector_type="mca",
                    data=live_time[i],
                    local_alias="",
                    global_alias=f"{group}_live_time",
                )


def _iter_xia_mca_data(zap_info: ZapInfo, detnr: int) -> Iterator[ScanData]:
    """We have one xiaXX file for each detector and each line in a 2D scan. The 2D data in
    each xiaXX file has the shape `(ncols, nchan)` where

        * `ncols` the number of points in each line of a 2D scan
        * `nchan`: the number of MCA channels
    """
    pattern = os.path.join(
        zap_info.directory,
        f"{zap_info.radix}_xia{detnr:02d}_{zap_info.scannb:04d}_{zap_info.zapnb:04d}_[0-9][0-9][0-9][0-9].edf",
    )
    files = sorted(glob(pattern))
    for filename in files:
        with fabio.open(filename) as f:
            yield ScanData(
                group=f"xia_det{detnr}",
                name="spectrum",
                detector_type="mca",
                data=f.data,
                local_alias="data",
                global_alias=f"xia_det{detnr}",
            )


def _iter_lima_data(zap_info: ZapInfo) -> Iterator[ScanData]:
    """We have one lima file for each detector and for each point in a 2D scan."""
    pattern = os.path.join(
        zap_info.directory,
        f"{zap_info.radix}_*_{zap_info.scannb:04d}_[0-9][0-9][0-9][0-9]_[0-9][0-9][0-9][0-9].edf",
    )
    files = {_detector_key(filename): filename for filename in glob(pattern)}
    for (camera, *_), filename in sorted(files.items()):
        if camera.startswith("xia"):
            continue
        with fabio.open(filename) as f:
            data = f.data
            if data.ndim != 2:
                continue
            yield ScanData(
                group=camera,
                name="image",
                detector_type="lima",
                data=data[numpy.newaxis, ...],
                local_alias="data",
                global_alias=camera,
            )


def _detector_key(filename: str) -> Tuple[str, int, int]:
    parts = os.path.splitext(os.path.basename(filename))[0].split("_")
    camera, _, slownr, fastnr = parts[-4:]
    return camera, int(slownr), int(fastnr)


def _iter_xia_ctr_data(
    zap_info: Dict[str, numpy.ndarray],
    exclude_mca_detectors: Sequence[int],
    motor_names: Sequence[str],
) -> Iterator[ScanData]:
    """We have one file for each counter with the shape `(nrows, ncols)` where

    * `nrows` the number of lines in a 2D scan
    * `ncols` the number of points in each line of a 2D scan
    """
    pattern = os.path.join(
        zap_info.directory,
        f"{zap_info.radix}_*_{zap_info.scannb:04d}_{zap_info.zapnb:04d}.edf",
    )
    files = glob(pattern)
    if not files:
        return

    before, _, after = pattern.partition("*")
    nbefore = len(before)
    nafter = len(after)
    raw_camera_names = set()

    for filename in files:
        raw_ctr_name = filename[nbefore:-nafter]
        if raw_ctr_name in raw_camera_names:
            continue
        is_xmap_counter = _XMAP_COUNTER_PATTERN.match(raw_ctr_name)
        if is_xmap_counter:
            roi, detnr = is_xmap_counter.groups()
            if detnr == "Sum":
                continue
            detnr = int(detnr)
            if detnr in exclude_mca_detectors:
                continue
            group = f"xia_det{detnr}"
            name = rename_counter(roi)
            detector_type = "mca"
            global_alias = f"{group}_{name}"
        else:
            group = rename_counter(raw_ctr_name)
            name = "data"
            global_alias = group
            if group in motor_names:
                detector_type = "positioner"
            else:
                detector_type = ""

        with fabio.open(filename) as f:
            if "ccd_expo_time" in f.header:
                raw_camera_names.add(raw_ctr_name)
                continue
            yield ScanData(
                group=group,
                name=name,
                detector_type=detector_type,
                data=f.data.flatten(),
                local_alias="",
                global_alias=global_alias,
            )


def _get_motors_from_xia_header(header: Dict[str, str]) -> Dict[str, float]:
    motors = dict()
    for name, value in header.items():
        if name in _REMOVE_EDF_HEADER_KEYS:
            continue
        try:
            value = float(value)
        except ValueError:
            continue
        motors[name] = value
    return motors



import logging
import numbers
from typing import Dict
from typing import Optional
from typing import Tuple

import numpy

from ..io import hdf5

_ENERGY_UNITS = {"kev": 1, "ev": 0.001}
_ENERGY_TEMPLATE = "instrument/positioners_start/{}"

_logger = logging.getLogger(__name__)


def get_primary_beam_energy_suburi(bliss_scan_uri: str) -> Optional[str]:
    return _get_unit_position_suburi(bliss_scan_uri, _ENERGY_UNITS)


def get_primary_beam_energy_value(
    bliss_scan_uri: str,
    energy_name: Optional[str] = None,
    energy_uri_template: Optional[str] = None,
    search_on_units: bool = False,
) -> Optional[float]:
    """Get the incident beam energy from a Bliss scan."""
    if energy_name:
        if not energy_uri_template:
            energy_uri_template = _ENERGY_TEMPLATE
        try:
            energy, units = _get_template_position_value(
                bliss_scan_uri, energy_name, energy_uri_template
            )
        except KeyError:
            _logger.warning(
                "'%s' does not exist. Do not use the incident beam energy from the data.",
                energy_name,
            )
            energy, units = None, None
    elif search_on_units:
        energy, units = _get_unit_position_value(bliss_scan_uri, _ENERGY_UNITS)
    else:
        energy, units = None, None

    if energy is None:
        _logger.warning("No incident beam energy position name is provided.")
        return None
    if units not in _ENERGY_UNITS:
        _logger.warning("Ignore unexpected incident beam energy unit %r.", units)
        return energy
    return energy * _ENERGY_UNITS[units]


def _get_unit_position_suburi(
    bliss_scan_uri: str, units: Dict[str, float]
) -> Optional[str]:
    """Get Bliss scan sub-URI for the first positioner with one of the units."""
    scan_filename, scan_h5path = hdf5.split_h5uri(bliss_scan_uri)
    with hdf5.FileReadAccess(scan_filename) as nxroot:
        positioners = nxroot[f"{scan_h5path}/instrument/positioners_start"]
        name = _get_positioner_name(positioners, units)
        if name is not None:
            return f"instrument/positioners_start/{name}"


def _get_template_position_value(
    bliss_scan_uri: str, position_name: str, position_uri_template: str
) -> Tuple[float, str]:
    """Get position value and unit from a Bliss scan."""
    scan_filename, scan_h5path = hdf5.split_h5uri(bliss_scan_uri)
    suburi = position_uri_template.format(position_name)
    with hdf5.FileReadAccess(scan_filename) as nxroot:
        data_path = f"{scan_h5path}/{suburi}"
        dset = nxroot[data_path]
        punits = dset.attrs.get("units", "").lower()
        return _asscalar(dset[()]), punits


def _get_unit_position_value(
    bliss_scan_uri: str, units: Dict[str, float]
) -> Optional[Tuple[numbers.Number, str]]:
    """Get the first positioner value from a Bliss scan with one of the units."""
    scan_filename, scan_h5path = hdf5.split_h5uri(bliss_scan_uri)
    with hdf5.FileReadAccess(scan_filename) as nxroot:
        positioners = nxroot[f"{scan_h5path}/instrument/positioners_start"]
        return _get_positioner_value(positioners, units)


def _get_positioner_value(
    positioners: hdf5.GroupType, units: Dict[str, float]
) -> Optional[Tuple[numbers.Number, str]]:
    """Get the first dataset value with one of the units."""
    for name in positioners:
        positioner = positioners[name]
        punits = positioner.attrs.get("units", "").lower()
        if punits in units:
            return _asscalar(positioner[()]), punits


def _asscalar(value: numpy.ndarray, rtol=1.0e-5, atol=1.0e-8) -> numbers.Number:
    if numpy.isscalar(value):
        return value

    if not isinstance(value, numpy.ndarray):
        raise TypeError(f"Unsupported type: {type(value)}")

    if value.size == 0:
        raise ValueError("Cannot convert empty array to scalar")
    if value.size == 1:
        return value.item()
    ref_value = numpy.nanmedian(value)
    if numpy.allclose(value, ref_value, atol=atol, rtol=rtol):
        return ref_value
    raise ValueError(f"Array values differ beyond tolerance (shape {value.shape})")


def _get_positioner_name(
    positioners: hdf5.GroupType, units: Dict[str, float]
) -> Optional[str]:
    for name in positioners:
        positioner = positioners[name]
        punits = positioner.attrs.get("units", "").lower()
        if punits in units:
            return name



"""Buffers receive data from pymca."""

from .abstract import AbstractOutputBuffer  # noqa F401
from .external import ExternalOutputBuffer  # noqa F401
from .pymca import PyMcaOutputBuffer  # noqa F401



"""Handlers receive data from buffers."""

from .abstract import AbstractOutputHandler  # noqa F401
from .nexus import NexusOutputHandler  # noqa F401
from .queue import QueueOutputHandler  # noqa F401
from .queue import consume_handler_queue  # noqa F401
from .queue import stop_queue  # noqa F401



import h5py
import pytest
from ewoksorange.tests.utils import execute_task

from orangecontrib.ewoksfluo.categories.fit.fit_multi_scan_multi_det import (
    OWFitStackMultiDetector,
)
from orangecontrib.ewoksfluo.categories.fit.fit_multi_scan_single_det import (
    OWFitStackSingleDetector,
)
from orangecontrib.ewoksfluo.categories.fit.fit_single_scan_multi_det import (
    OWFitSingleScanMultiDetector,
)
from orangecontrib.ewoksfluo.categories.fit.fit_single_scan_single_det import (
    OWFitSingleScanSingleDetector,
)

from .. import utils

TASK_CLASSES_BY_SHAPE = {
    (1, 1): OWFitSingleScanSingleDetector,
    (1, 2): OWFitSingleScanMultiDetector,
    (2, 1): OWFitStackSingleDetector,
    (2, 2): OWFitStackMultiDetector,
}


@pytest.mark.parametrize("nscans", [1, 2])
@pytest.mark.parametrize("ndetectors", [1, 2])
def test_fit_tasks(tmp_path, nscans, ndetectors):
    _test_example_xrf_scan(tmp_path, nscans, ndetectors)


@pytest.mark.parametrize("nscans", [1, 2])
@pytest.mark.parametrize("ndetectors", [1, 2])
def test_fit_tasks_widget(tmp_path, qtapp, nscans, ndetectors):
    _test_example_xrf_scan(tmp_path, nscans, ndetectors, widget=True)


def _test_example_xrf_scan(tmp_path, nscans, ndetectors, widget: bool = False):
    task_cls = TASK_CLASSES_BY_SHAPE[(nscans, ndetectors)]
    if not widget:
        task_cls = task_cls.ewokstaskclass

    _, _, _, config = utils.generate_data(
        10, 8.5, tmp_path=tmp_path, nscans=nscans, ndetectors=ndetectors
    )

    filename = str(tmp_path / "output.h5")
    inputs = {
        "quantification": True,
        "fast_fitting": True,
        "diagnostics": True,
        "xrf_spectra_uri_template": "measurement/{}",
    }

    if nscans == 1 and ndetectors == 1:
        inputs["bliss_scan_uri"] = f"{str(tmp_path / 'spectra.h5')}::/1.1"
        inputs["output_root_uri"] = f"{filename}::/1.1"
        inputs["detector_name"] = "mca00"
        inputs["config"] = config
    if nscans > 1 and ndetectors == 1:
        inputs["bliss_scan_uris"] = [
            f"{str(tmp_path / 'spectra.h5')}::/{scan}.1"
            for scan in range(1, nscans + 1)
        ]
        inputs["output_root_uri"] = f"{filename}::/1.1"
        inputs["detector_name"] = "mca00"
        inputs["config"] = config
    if nscans == 1 and ndetectors > 1:
        inputs["bliss_scan_uri"] = f"{str(tmp_path / 'spectra.h5')}::/1.1"
        inputs["output_root_uri"] = f"{filename}::/1.1"
        inputs["detector_names"] = [
            f"mca{detector:02d}" for detector in range(ndetectors)
        ]
        inputs["configs"] = [config] * ndetectors
    if nscans >= 1 and ndetectors > 1:
        inputs["bliss_scan_uris"] = [
            f"{str(tmp_path / 'spectra.h5')}::/{scan}.1"
            for scan in range(1, nscans + 1)
        ]
        inputs["output_root_uri"] = f"{filename}::/1.1"
        inputs["detector_names"] = [
            f"mca{detector:02d}" for detector in range(ndetectors)
        ]
        inputs["configs"] = [config] * ndetectors

    for _ in range(2):  # Repeat twice to test overwrite
        outputs = execute_task(task_cls, inputs)
        expected = expected_task_outputs(nscans, ndetectors, tmp_path, inputs)
        assert outputs == expected

        with h5py.File(filename) as f:
            content = utils.h5content(f)

        expected = expected_h5content(nscans, ndetectors)
        assert content == expected


def expected_task_outputs(nscans, ndetectors, tmp_path, inputs):
    filename = str(tmp_path / "output.h5")
    outputs = dict()

    if ndetectors == 1:
        detector_name = inputs["detector_name"]
        outputs["xrf_results_uri"] = f"{filename}::/1.1/fit/{detector_name}/results"
        outputs["detector_name"] = detector_name
        outputs["output_root_uri"] = f"{filename}::/1.1"
    else:
        outputs["xrf_results_uris"] = [
            f"{filename}::/1.1/fit/mca{detector:02d}/results"
            for detector in range(ndetectors)
        ]
        outputs["detector_names"] = inputs["detector_names"]
        outputs["output_root_uri"] = f"{filename}::/1.1"

    if nscans == 1:
        outputs["bliss_scan_uri"] = inputs["bliss_scan_uri"]
    else:
        outputs["bliss_scan_uris"] = inputs["bliss_scan_uris"]

    return outputs


def expected_h5content(nscans, ndetectors):
    content = {"@attrs": {"NX_class", "default"}}
    fit = {"@attrs": {"NX_class", "default"}}
    entry = {"@attrs": {"NX_class", "default"}, "fit": fit}
    if nscans > 1:
        scan_shape = (2,)
    else:
        scan_shape = tuple()
    content["1.1"] = entry

    nxprocess_content = {
        "@attrs": {"NX_class", "default"},
        "configuration": {
            "@attrs": {"NX_class"},
            "data@shape": (),
            "date@shape": (),
            "type@shape": (),
        },
        "program@shape": (),
        "version@shape": (),
        "results": {
            "@attrs": {"NX_class", "default"},
            "derivatives": {
                "@attrs": {"NX_class", "auxiliary_signals", "axes", "signal"},
                "Al_K@shape": scan_shape + (1024,),
                "Ca_K@shape": scan_shape + (1024,),
                "Ce_L@shape": scan_shape + (1024,),
                "Cl_K@shape": scan_shape + (1024,),
                "Fe_K@shape": scan_shape + (1024,),
                "P_K@shape": scan_shape + (1024,),
                "Pb_M@shape": scan_shape + (1024,),
                "S_K@shape": scan_shape + (1024,),
                "Scatter_Compton000@shape": scan_shape + (1024,),
                "Scatter_Peak000@shape": scan_shape + (1024,),
                "Si_K@shape": scan_shape + (1024,),
                "Ti_K@shape": scan_shape + (1024,),
                "energy@attrs": {"units"},
                "energy@shape": (1024,),
            },
            "diagnostics": {
                "@attrs": {"NX_class", "auxiliary_signals", "signal"},
                "nFreeParameters@shape": scan_shape + (10,),
                "nObservations@shape": scan_shape + (10,),
            },
            "fit": {
                "@attrs": {"NX_class", "auxiliary_signals", "axes", "signal"},
                "data@shape": scan_shape + (10, 1024),
                "energy@attrs": {"units"},
                "energy@shape": (1024,),
                "model@shape": scan_shape + (10, 1024),
                "residuals@shape": scan_shape + (10, 1024),
            },
            "massfractions": {
                "@attrs": {"NX_class", "auxiliary_signals", "signal"},
                "Al_K@shape": scan_shape + (10,),
                "Ca_K@shape": scan_shape + (10,),
                "Ce_L@shape": scan_shape + (10,),
                "Cl_K@shape": scan_shape + (10,),
                "Fe_K@shape": scan_shape + (10,),
                "P_K@shape": scan_shape + (10,),
                "Pb_M@shape": scan_shape + (10,),
                "S_K@shape": scan_shape + (10,),
                "Si_K@shape": scan_shape + (10,),
                "Ti_K@shape": scan_shape + (10,),
            },
            "parameters": {
                "@attrs": {"NX_class", "auxiliary_signals", "signal"},
                "Al_K@shape": scan_shape + (10,),
                "Ca_K@shape": scan_shape + (10,),
                "Ce_L@shape": scan_shape + (10,),
                "Cl_K@shape": scan_shape + (10,),
                "Fe_K@shape": scan_shape + (10,),
                "P_K@shape": scan_shape + (10,),
                "Pb_M@shape": scan_shape + (10,),
                "S_K@shape": scan_shape + (10,),
                "Scatter_Compton000@shape": scan_shape + (10,),
                "Scatter_Peak000@shape": scan_shape + (10,),
                "Si_K@shape": scan_shape + (10,),
                "Ti_K@shape": scan_shape + (10,),
            },
            "uncertainties": {
                "@attrs": {"NX_class", "auxiliary_signals", "signal"},
                "Al_K@shape": scan_shape + (10,),
                "Ca_K@shape": scan_shape + (10,),
                "Ce_L@shape": scan_shape + (10,),
                "Cl_K@shape": scan_shape + (10,),
                "Fe_K@shape": scan_shape + (10,),
                "P_K@shape": scan_shape + (10,),
                "Pb_M@shape": scan_shape + (10,),
                "S_K@shape": scan_shape + (10,),
                "Scatter_Compton000@shape": scan_shape + (10,),
                "Scatter_Peak000@shape": scan_shape + (10,),
                "Si_K@shape": scan_shape + (10,),
                "Ti_K@shape": scan_shape + (10,),
            },
        },
    }
    for detector in range(ndetectors):
        fit[f"mca{detector:02d}"] = nxprocess_content
    return content



from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import Field
from pydantic import PrivateAttr


class PyMcaBaseModel(BaseModel, extra="allow", validate_assignment=True):
    pass


EnergyType = Union[float, None, Literal["None", "NONE"]]


class Fit(PyMcaBaseModel):
    deltaonepeak: Optional[float] = None
    strategy: Optional[str] = None
    strategyflag: Optional[int] = None
    fitfunction: Optional[int] = None
    continuum: Optional[int] = None
    fitweight: Optional[int] = None
    stripalgorithm: Optional[int] = Field(default=None, description="0: STRIP, 1: SNIP")
    linpolorder: Optional[int] = None
    exppolorder: Optional[int] = None
    stripconstant: Optional[float] = None
    snipwidth: Optional[int] = None
    stripiterations: Optional[int] = None
    stripwidth: Optional[int] = None
    stripfilterwidth: Optional[int] = None
    stripanchorsflag: Optional[int] = None
    maxiter: Optional[int] = None
    deltachi: Optional[float] = None
    xmin: Optional[int] = None
    xmax: Optional[int] = None
    linearfitflag: Optional[int] = None
    use_limit: Optional[int] = None
    stripflag: Optional[int] = None
    escapeflag: Optional[int] = None
    sumflag: Optional[int] = None
    scatterflag: Optional[int] = None
    hypermetflag: Optional[int] = None

    stripanchorslist: List[int] = Field(default_factory=list)

    energy: Union[List[EnergyType], EnergyType] = Field(default_factory=list)
    energyweight: Union[List[float], float] = Field(default_factory=list)
    energyflag: Union[List[int], int] = Field(default_factory=list)
    energyscatter: Union[List[int], int] = Field(default_factory=list)


class Concentrations(PyMcaBaseModel):
    usematrix: Optional[int] = None
    useattenuators: Optional[int] = None
    usemultilayersecondary: Optional[int] = None
    usexrfmc: Optional[int] = None
    mmolarflag: Optional[int] = None
    flux: Optional[float] = None
    time: Optional[float] = None
    area: Optional[float] = None
    distance: Optional[float] = None
    reference: Optional[str] = None
    useautotime: Optional[int] = None


class Detector(PyMcaBaseModel):
    detele: str
    nthreshold: int

    zero: float
    deltazero: float
    fixedzero: int

    gain: float
    deltagain: float
    fixedgain: int

    noise: float
    deltanoise: float
    fixednoise: int

    fano: float
    deltafano: float
    fixedfano: int

    sum: float
    deltasum: float
    fixedsum: int

    ignoreinputcalibration: Optional[int] = None

    ethreshold: Optional[float] = None
    ithreshold: Optional[float] = None


class PeakShape(PyMcaBaseModel):
    st_arearatio: float
    deltast_arearatio: float
    fixedst_arearatio: int

    st_sloperatio: float
    deltast_sloperatio: float
    fixedst_sloperatio: int

    lt_arearatio: float
    deltalt_arearatio: float
    fixedlt_arearatio: int

    lt_sloperatio: float
    deltalt_sloperatio: float
    fixedlt_sloperatio: int

    step_heightratio: float
    deltastep_heightratio: float
    fixedstep_heightratio: int

    eta_factor: Optional[float] = None
    deltaeta_factor: Optional[float] = None
    fixedeta_factor: Optional[int] = None


class Material(PyMcaBaseModel):
    Comment: Optional[str] = None
    Density: float
    Thickness: Optional[float] = None
    CompoundFraction: Union[float, Literal[""], List[float]]
    CompoundList: Union[str, List[str]] = Field(
        ..., description="key in materials or chemical formula"
    )


class UserAttenuator(PyMcaBaseModel):
    use: int
    name: str
    comment: Optional[str]
    energy: Union[float, List[float]]
    transmission: Union[float, List[float]]


class Tube(PyMcaBaseModel):
    transmission: int
    voltage: float
    anode: str
    anodethickness: float
    anodedensity: float
    window: str
    windowthickness: float
    windowdensity: float
    filter1: str
    filter1thickness: float
    filter1density: float
    alphax: float
    alphae: float
    deltaplotting: float


class XRFMCSetup(PyMcaBaseModel):
    p_polarisation: float
    source_sample_distance: float
    slit_distance: float
    slit_width_x: float
    slit_width_y: float
    source_size_x: float
    source_size_y: float
    source_diverg_x: float
    source_diverg_y: float
    nmax_interaction: int
    layer: int
    collimator_height: float
    collimator_diameter: float
    histories: int


class XRFMC(PyMcaBaseModel):
    program: Optional[str] = None
    setup: XRFMCSetup


class Fisx(PyMcaBaseModel):
    pass


class AttenuatorView(BaseModel):
    enabled: int
    material: Optional[str] = Field(
        ..., description="key in materials or chemical formula"
    )
    density: float = Field(..., description="g/cm^3")
    thickness: float = Field(..., description="cm")
    incidence_angle: Optional[float] = Field(
        None, description="Angle between incident beam and sample surface (degrees)"
    )
    exit_angle: Optional[float] = Field(
        None, description="Angle between outgoing beam and sample surface (degrees)"
    )
    scattering_flag: Optional[bool] = None
    scattering_angle: Optional[float] = Field(
        None, description="Angle between incident beam and outgoing beam (degrees)"
    )

    _raw: List[Any] = PrivateAttr()

    @classmethod
    def from_list(cls, data: List[Any]) -> "AttenuatorView":
        padded = (data + [None] * 8)[:8]
        obj = cls(
            enabled=padded[0],
            material=padded[1],
            density=padded[2],
            thickness=padded[3],
            incidence_angle=padded[4],
            exit_angle=padded[5],
            scattering_flag=padded[6],
            scattering_angle=padded[7],
        )
        obj._raw = data
        return obj

    def _sync(self) -> None:
        new = [
            self.enabled,
            self.material,
            self.density,
            self.thickness,
            self.incidence_angle,
            self.exit_angle,
            self.scattering_flag,
            self.scattering_angle,
        ]
        self._raw[:] = new[: len(self._raw)]

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name in type(self).model_fields:
            if hasattr(self, "_raw"):
                self._sync()


class MultilayerView(BaseModel):
    use: int
    material: str = Field(..., description="key in materials or chemical formula")
    density: float
    thickness: float

    _raw: List[Any] = PrivateAttr()

    @classmethod
    def from_list(cls, data: List[Any]) -> "MultilayerView":
        padded = (data + [None] * 4)[:4]
        obj = cls(
            use=padded[0],
            material=padded[1],
            density=padded[2],
            thickness=padded[3],
        )
        obj._raw = data
        return obj

    def _sync(self) -> None:
        self._raw[:] = [
            self.use,
            self.material,
            self.density,
            self.thickness,
        ]

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name in type(self).model_fields:
            if hasattr(self, "_raw"):
                self._sync()


class PyMcaXrfConfiguration(PyMcaBaseModel):
    peaks: Dict[str, Union[str, List[str]]] = Field(default_factory=dict)
    fit: Fit = Field(default_factory=dict)
    concentrations: Concentrations = Field(default_factory=dict)
    detector: Detector = Field(default_factory=dict)
    peakshape: PeakShape = Field(default_factory=dict)
    materials: Dict[str, Material] = Field(default_factory=dict)
    attenuators: Dict[str, List[Any]] = Field(default_factory=dict)
    multilayer: Dict[str, List[Any]] = Field(default_factory=dict)
    userattenuators: Dict[str, UserAttenuator] = Field(default_factory=dict)
    tube: Optional[Tube] = None
    xrfmc: Optional[XRFMC] = None
    fisx: Optional[Fisx] = None

    @property
    def attenuator_views(self) -> Dict[str, AttenuatorView]:
        return {
            name: AttenuatorView.from_list(values)
            for name, values in self.attenuators.items()
        }

    @property
    def multilayer_views(self) -> Dict[str, MultilayerView]:
        return {
            name: MultilayerView.from_list(values)
            for name, values in self.multilayer.items()
        }



from typing import Dict
from typing import List

from ewokscore import Task
from ewokscore.model import BaseInputModel
from ewokscore.model import BaseOutputModel
from pydantic import Field

from ...io.blissconcat import concatenate_bliss_scans


class Inputs(BaseInputModel):
    bliss_scan_uris: List[str] = Field(
        description="List of Bliss scan URI's to concatenate.",
        examples=[
            [
                "/data/dataset1.h5::/2.1",
                "/data/dataset1.h5::/4.1",
                "/data/dataset2.h5::/100.1",
                "/data/dataset2.h5::/101.1",
            ]
        ],
    )
    bliss_scan_uri: str = Field(
        description="Concatenated Bliss scan URI to save.",
        examples=[
            "/result/concatenated.h5",
            "/result/concatenated.h5::/2.1",
        ],
    )
    virtual_axes: Dict[str, str] = Field(
        default_factory=dict,
        description="Virtual positioners which are typically the sum of scalar and moving motor positions.",
        examples=[{"sy": "<nsy>+<nspy>", "sz": "<nsz>+<nspz>"}],
    )
    axes_units: Dict[str, str] = Field(
        default_factory=dict,
        description="Axes units to be used when missing.",
        examples=[{"nsy": "mm", "nsz": "mm", "nspy": "um", "nspz": "um"}],
    )


class Outputs(BaseOutputModel):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/result/concatenated.h5::/2.1"]
    )


class ConcatBliss(Task, input_model=Inputs, output_model=Outputs):
    """Concatenate Bliss scans."""

    def run(self):
        self.outputs.bliss_scan_uri = concatenate_bliss_scans(
            self.inputs.bliss_scan_uris,
            self.inputs.bliss_scan_uri,
            virtual_axes=self.inputs.virtual_axes,
            axes_units=self.inputs.axes_units,
        )



from contextlib import contextmanager
from typing import Any
from typing import Dict
from typing import Generator
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

import numpy
import pytest

from ...xrffit.fit import fit_xrf_spectra
from ...xrffit.fit.types import NotAvailable
from ...xrffit.fit.types import XRFBatchFitResult
from ..utils import generate_data


@pytest.mark.parametrize("fast_fitting", [False, True], ids=["slow", "fast"])
@pytest.mark.parametrize("use_limit", [False, True], ids=["wo_limit", "w_limit"])
@pytest.mark.parametrize(
    "quantification",
    [None, "fundamental", "auto", "ref"],
    ids=["quant_no", "quant_fundamental", "quant_auto", "quant_ref"],
)
@pytest.mark.parametrize(
    "positive_peak_areas", [False, True], ids=["no_refit", "refit"]
)
@pytest.mark.parametrize(
    "error_propagation",
    [None, "average", "individual"],
    ids=["error_no", "error_avg", "error_ind"],
)
@pytest.mark.parametrize(
    "strip_background", [None, 0, 1], ids=["bkg_no", "bkg_strip", "bkg_snip"]
)
@pytest.mark.parametrize("live_times", [False, True], ids=["no_LT", "LT"])
@pytest.mark.parametrize(
    "multilayer", [False, True], ids=["single_layer", "multilayer"]
)
@pytest.mark.parametrize("legacy", [False, True], ids=["pymca5", "pymca5_legacy"])
def test_fit_versus_native_fit(
    fast_fitting: bool,
    use_limit: bool,
    quantification: Optional[str],
    positive_peak_areas: bool,
    error_propagation: Optional[str],
    strip_background: Optional[Literal[0, 1]],
    live_times: bool,
    multilayer: bool,
    legacy: bool,
):
    """
    :param fast_fitting: fast fitting (forced linear) or slow fitting
    :param use_limit: fit the entire XRF spectrum or a ROI
    :param quantification: calulate concentrations from fitted peak areas
    :param positive_peak_areas: constraint peak areas to be `>=0`
    :param error_propagation: no error propagation, Poisson noise from average spectrum, Poisson noise for each spectrum separately)
    :param strip_background: no background stripping, strip, snip
    :param live_times: include live time normalization
    :param multilayer: the sample matrix is a "multilayer"
    :param legacy: legacy high-level native PyMca5 API for fitting
    """
    if not fast_fitting:
        if not positive_peak_areas:
            pytest.skip("Slow fitting always needs positive constraints")
        if live_times:
            pytest.skip("Slow fitting does not have live-time normalization integrated")
        if error_propagation == "average":
            pytest.skip("Slow fitting always needs individual weights")

    if strip_background == 0 and fast_fitting:
        expected_error = NotImplementedError
        expected_native_error = RuntimeError
    else:
        expected_error = None
        expected_native_error = None

    if multilayer:
        num_layers = 2
    else:
        num_layers = 1

    with _raises(expected_native_error) as has_native_result:
        native_results = _fit_xrf_spectra(
            fast_fitting=fast_fitting,
            use_limit=use_limit,
            quantification=quantification,
            error_propagation=error_propagation,
            positive_peak_areas=positive_peak_areas,
            strip_background=strip_background,
            live_times=live_times,
            native_fitting=True,
            native_legacy_fitting=legacy,
            num_layers=num_layers,
        )

    with _raises(expected_error) as has_result:
        results = _fit_xrf_spectra(
            fast_fitting=fast_fitting,
            use_limit=use_limit,
            quantification=quantification,
            error_propagation=error_propagation,
            positive_peak_areas=positive_peak_areas,
            strip_background=strip_background,
            live_times=live_times,
            native_fitting=False,
            native_legacy_fitting=False,
            num_layers=num_layers,
        )

    assert has_result == has_native_result
    if not has_result:
        return

    if use_limit:
        _assert_channel_size(results)
        _assert_channel_size(native_results)

    results, compare_options = _prepare_native_compare_fitresults(
        results, native_results, fast_fitting, native_results
    )
    _compare_data(results, native_results, **compare_options)


def _assert_channel_size(results: XRFBatchFitResult) -> None:
    names = ["channels", "energies"]
    for name in names:
        if results.has_data(name):
            assert getattr(results, name).shape[-1] == 1021, name

    names = [
        "raw_channels",
        "raw_energies",
        "derivatives",
        "spectra",
        "model",
        "residuals",
    ]
    for name in names:
        if results.has_data(name):
            assert getattr(results, name).shape[-1] == 1024, name


def _prepare_native_compare_fitresults(
    results: XRFBatchFitResult,
    native_results: XRFBatchFitResult,
    fast_fitting: bool,
    legacy: bool,
) -> Tuple[XRFBatchFitResult, Dict[str, Union[float, bool, None]]]:
    if legacy:
        # Stores results in float32
        compare_options = dict(float_atol=1e-5, float_rtol=1e-2, float_as_32bit=True)
    else:
        compare_options = dict(float_atol=None, float_rtol=None, float_as_32bit=False)

    if not fast_fitting:
        # Native slow fitting does not send non-linear fit parameters to the output buffer
        new_names = results.parameter_names
        native_names = native_results.parameter_names
        idx = [i for i, name in enumerate(new_names) if name in native_names]
        results = type(results)(
            **{
                **results._asdict(),
                "parameter_names": native_names,
                "parameters": results.parameters[idx],
                "uncertainties": results.uncertainties[idx],
            }
        )

        # Stores results in float32
        compare_options["float_as_32bit"] = True
        compare_options["float_rtol"] = 1e-1  # TODO: 10% is too much

    return results, compare_options


def _fit_xrf_spectra(
    fast_fitting: bool,
    use_limit: bool,
    quantification: Optional[str],
    error_propagation: str,
    positive_peak_areas: bool,
    strip_background: Optional[str],
    live_times: bool,
    native_fitting: bool,
    native_legacy_fitting: bool,
    num_layers: int,
) -> numpy.ndarray:
    if positive_peak_areas:
        points_with_negative_peaks = (1,)
    else:
        points_with_negative_peaks = tuple()

    if error_propagation is None:
        fitweight = False
        individual_weights = None
    else:
        fitweight = True
        if error_propagation == "individual":
            individual_weights = True
        elif error_propagation == "average":
            individual_weights = False
        else:
            raise ValueError(error_propagation)

    if strip_background is None:
        stripflag = False
        stripalgorithm = 0  # Does not matter
    else:
        stripflag = True
        stripalgorithm = strip_background

    _, xrf_spectra, _, configuration = generate_data(
        npoints_per_scan=2,
        energy=7.5,
        points_with_negative_peaks=points_with_negative_peaks,
        quantification=quantification,
        num_layers=num_layers,
    )
    configuration = configuration.model_copy(deep=True)
    xrf_spectra = xrf_spectra[0]

    configuration.fit.use_limit = int(use_limit)
    configuration.fit.fitweight = int(fitweight)
    configuration.fit.stripflag = int(stripflag)
    configuration.fit.stripalgorithm = int(stripalgorithm)

    if live_times:
        # time in pymca config is 1 sec
        live_times = numpy.full((len(xrf_spectra),), 0.5)
        configuration.concentrations.useautotime = 1
    else:
        live_times = None
        configuration.concentrations.useautotime = 0

    original_xrf_spectra = xrf_spectra.copy()
    results = fit_xrf_spectra(
        xrf_spectra,
        configuration,
        quantification=bool(quantification),
        individual_weights=individual_weights,
        live_times=live_times,
        positive_peak_areas=positive_peak_areas,
        diagnostics=True,
        fast_fitting=fast_fitting,
        native_fitting=native_fitting,
        native_legacy_fitting=native_legacy_fitting,
    )
    numpy.testing.assert_array_equal(xrf_spectra, original_xrf_spectra)
    return results


@contextmanager
def _raises(exc_cls: Optional[Type[Exception]]) -> Generator[bool, None, None]:
    try:
        if exc_cls:
            with pytest.raises(exc_cls):
                yield False
        else:
            yield True
    except NotImplementedError as ex:
        pytest.skip(str(ex))


def _compare_data(
    actual: Any,
    desired: Any,
    _name: Optional[str] = "<root>",
    float_atol: Optional[float] = None,
    float_rtol: Optional[float] = None,
    float_as_32bit: bool = False,
) -> None:
    err_msg = f"{_name!r} is different"

    if isinstance(desired, NotAvailable):
        return
    assert type(actual) is type(desired), err_msg

    if isinstance(desired, dict):
        for k, v in desired.items():
            _compare_data(
                actual[k],
                v,
                _name=f"{_name}.{k}",
                float_atol=float_atol,
                float_rtol=float_rtol,
                float_as_32bit=float_as_32bit,
            )
    elif isinstance(desired, numpy.ndarray):
        assert actual.shape == desired.shape, err_msg

        atol = 1e-10
        rtol = 1e-7

        if numpy.issubdtype(desired.dtype, numpy.floating):
            if float_atol is not None:
                atol = float_atol
            if float_rtol is not None:
                rtol = float_rtol
            if float_as_32bit:
                with numpy.errstate(over="ignore"):  # Stores results in float32
                    actual = actual.astype(numpy.float32)
                    desired = desired.astype(numpy.float32)

        # abs(actual - desired) < atol + rtol * abs(desired)
        numpy.testing.assert_allclose(
            actual, desired, err_msg=err_msg, atol=atol, rtol=rtol
        )

    elif isinstance(desired, (tuple, list)):
        assert len(actual) == len(desired), err_msg
        if hasattr(desired, "_fields"):
            names = desired._fields
        else:
            names = list(range(len(actual)))
        for name, v1, v2 in zip(names, actual, desired):
            _compare_data(
                v1,
                v2,
                _name=f"{_name}[{name}]",
                float_atol=float_atol,
                float_rtol=float_rtol,
                float_as_32bit=float_as_32bit,
            )
    else:
        assert actual == desired, err_msg
